(*
This file is part of the Kind verifier

* Copyright (c) 2007-2012 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


(** Test case generator.  Main entry point. *)

(** 
@author Temesghen Kahsai

*)

open Types
open Exceptions
open Channels
open Globals


let solver = Globals.my_solver

(********************************************************************)
let print_copyright () =
  print_string (
   "\n Test Case Generator based on BMC (version "^Globals.version
  ^", build "^Globals.build^")\n"
  ^"* Copyright (c) 2007-2012 by the Board of Trustees of the University of Iowa.\n"
  ^"* All rights reserved.\n"
  ^"* \n"
  ^"* Redistribution and use in source and binary forms, with or without\n"
  ^"* modification, are permitted provided that the following conditions are met:\n"
  ^"*     * Redistributions of source code must retain the above copyright\n"
  ^"*       notice, this list of conditions and the following disclaimer.\n"
  ^"*     * Redistributions in binary form must reproduce the above copyright\n"
  ^"*       notice, this list of conditions and the following disclaimer in the\n"
  ^"*       documentation and/or other materials provided with the distribution.\n"
  ^"*     * Neither the name of the University of Iowa, nor the\n"
  ^"*       names of its contributors may be used to endorse or promote products\n"
  ^"*       derived from this software without specific prior written permission.\n"
  ^"*\n"
(*  
  ^"* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY\n"
  ^"* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n"
  ^"* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n"
  ^"* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY\n"
  ^"* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n"
  ^"* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;\n"
  ^"* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND\n"
  ^"* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n"
  ^"* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS\n"
  ^"* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n"
*)
  )


(********************************************************************)
let catch_alarm x =
  print_to_user_final (solver#get#cc^"Signal caught: Timeout. Terminating.\n");
  exit x

let catch_term x =
  print_to_user_final (solver#get#cc^"Signal caught: Memory out. Terminating.\n");
  exit x

let catch_pipe x = 
  print_to_user_final (solver#get#cc^"Signal caught: Broken pipe. The solver has unexpectedly closed.  Terminating.\n"
  ^(match !Flags.set_my_solver with
        YICES -> ""
      | YICES_WRAPPER -> 
          solver#get#cc^"It is possible a formula exceeded the interactive yices buffer.\n"
      | CVC3 -> (if !Flags.abs_mode then 
                   solver#get#cc^"It is possible CVC3 failed to instantiate a counterexample.\n"
                   ^solver#get#cc^"Try re-running with abstraction turned off.\n" else "")
      | Z3 -> ""
      | CVC4 -> (if !Flags.abs_mode then 
                   solver#get#cc^"It is possible CVC4 failed to instantiate a counterexample.\n"
                   ^solver#get#cc^"Try re-running with abstraction turned off.\n" else ""))
		      );
  exit x
			 


(********************************************************************)

let start fn =
  begin
    print_to_user_final ("\n\n==========================================\n");
    print_to_user_final("  KIND TEST CASE GENERATION               \n");
    print_to_user_final ("\n==========================================\n");
    match !Flags.set_my_solver with
        YICES ->

          solver#set (new Solver_yices_cmdline.solver_yices_cmdline)
      | YICES_WRAPPER -> 
	  begin
	    (* In case the yices warpper is not available switch to yices command line version *)
	    if(Solvers_path.yicesw_path = "no") then
	      (
		solver#set (new Solver_yices_cmdline.solver_yices_cmdline)
	      )else(
		solver#set (new Solver_yices.solver_yices)
	      );
	  end            
      | CVC3 -> 
          solver#set (new Solver_cvc3.solver_cvc3)

      | CVC4 -> 
	  solver#set (new Solver_cvc4.solver_cvc4)

      | Z3 ->
	  solver#set (new Solver_z3.solver_z3)
      end;

  (* Store the commmands sent to the solver in auxilary files, i.e. for debugging purposes.*)
      if !Flags.do_scratch then(
         base_ch := open_out (fn^"."^solver#get#file_extension^"_tg")
      ); 

      (* If XML is activated generate the output in a file*)
      if !Flags.xml then (
	 print_to_user_final ("  KIND results will be saved in: " ^ fn ^ ".xml\n");
	xml_ch := open_out (fn^".xml");
	print_xml (Kind_support.print_header_xml fn);
      );

Bmc_test_gen.main fn;;
  

(********************************************************************)
let _ =
  let usage_msg = "kind-test-gen [OPTIONS] [LUSTRE_FILE]" in
  let speclist = [
		  
                   ("-node", Arg.String (fun x -> Flags.user_specified_main_node_name := x),
		     "\tCONTROL: User specified main node (default: the last node that contains a PROPERTY)");

                   ("-n",
                    Arg.Int (fun x -> Flags.maxloops := x),
                    "\tCONTROL: Stop after n iterations (default 200)"
                   );

		   ("-no_multi",
                    Arg.Unit (fun () -> Flags.no_multi := true),
                    "\tCONTROL: Disables the verification of multiple properties in an incremental fascion. (default: true)"
		   );

		   ("-timeout",
                    Arg.Float (fun x -> Flags.timeout := x),
                    "\tCONTROL: For setting a timeout (default: 100.00)"
                   );
         		   		   
		   ("-yices_exec",
                    Arg.Unit (fun () -> Flags.set_my_solver := YICES),
                    "\tSOLVER: Use Yices executable as solver (default: yices)"
                   );
  
		   
		   ("-flags",
                    Arg.String (fun x -> Flags.solverflags := x),
                    "\tSOLVER: Pass string to solver cammand line (default: n/a)"
                   );

		   ("-xml",
		    Arg.Unit (fun x -> Flags.xml := true),
                    "\tOUTPUT: Generate KIND result in XML format (default: false)"
                   );
		   
        ("-quiet",
                    Arg.Unit (fun () -> Flags.loud := false;
                                        Flags.final_cex_loud := false),
                    "\tOUTPUT: Only return final answer (default: true)"
                   );
                   
		   ("-verbose",
                    Arg.Unit (fun () -> Flags.loud := true;
                                        Flags.final_cex_loud := true),
                    "\tOUTPUT: Return more details (default: false)"
                   );

		 
		    ("-scratch",
                    Arg.Unit (fun () -> Flags.do_scratch := true),
                    "\tOUTPUT: Produce files for debugging purposes. (default: Don't log to disk)"
                   );
    
        ("-version",
                      Arg.Unit (fun () -> print_copyright()),
                      "Print version & copyright information"
                   )
                 ]
  in
  Arg.parse speclist start usage_msg 


