(*
This file is part of the Kind verifier

* Copyright (c) 2007-2011 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(** A logging system for Kind-2 *)

(** 

    @author Corey Oliver
    @author Temesghen Kahsai

*)


open Printf

type level = SEVERE | WARNING | INFO | DEBUG
type formatter = out_channel -> level -> float -> string -> unit


type logger = {
  mutable level : level;
  mutable process : Types.proc
  mutable formatter : formatter;
  mutable channel : out_channel;
  mutable write_to_file : bool;
  mutable file_mode : open_flag list;
  mutable file_perm : int;
  mutable file_name : string;
}

let level_to_string = function
  | SEVERE -> "SEVERE"
  | WARNING -> "WARNING"
  | INFO -> "INFO"
  | DEBUG -> "DEBUG"

let process_to_string = function
  | BASE_PROC -> "BASE_PROC"
  | INDUCT_PROC -> "INDUCT_PROC"
  | INV_GEN_PROC -> "INV_GEN_PROC"

let format_timestamp out ts =
  let tm = Unix.gmtime ts in
  let ms,_ = modf ts in
  fprintf out "%04d-%02d-%02dT%02d:%02d:%02d.%06dZ"
    (1900 + tm.Unix.tm_year)
    (1 + tm.Unix.tm_mon)
    (tm.Unix.tm_mday)
    (tm.Unix.tm_hour)
    (tm.Unix.tm_min)
    (tm.Unix.tm_sec)
    (int_of_float (100000. *. ms))
    
(******************************************************************************)

let std_formatter channel level timestamp message =
  fprintf channel "D:%a\tL:%s\t%s\n%!"
    format_timestamp timestamp
    (level_to_string level)
    message
  
let mk_logger level process formatter channel write_to_file file_mode file_perm file_name =
  { level = level;
    process = process;
    formatter = formatter;
    channel = channel;
    write_to_file = write_to_file;
    file_mode = file_mode;
    file_perm = file_perm;
    file_name = file_name }

let log
    logger
    ?(level=logger.level)
    ?(process=logger.process)
    ?(timestamp=Unix.time () )
    message =
  logger.formatter
    logger.channel
    level
    timestamp
    message;
  if logger.write_to_file then
    logger.formatter
      (open_out_gen logger.file_mode
	 logger.file_perm logger.file_name)
      level
      timestamp
      message

let set_level logger level = logger.level <- level

let set_proc logger process = logger.process <- process
    
let set_formatter logger formatter = logger.formatter <- formatter
    
let set_channel logger channel = logger.channel <- channel
    
let set_write_to_file logger write_to_file = logger.write_to_file <- write_to_file

let set_file_mode logger file_mode = logger.file_mode <- file_mode

let set_file_perm logger file_perm = logger.file_perm <- file_perm

let set_file_name logger file_name = logger.file_name <- file_name

(******************************************************************************)

let test () =
  let l = (mk_logger BASE_PROC SEVERE std_formatter stdout true [Open_append] 777 "log.txt") in
  log l "hello";
  log l ~level:SEVERE "severe msg";
  log l ~level:WARNING "warning msg";
  log l ~level:INFO "info msg";
  log l ~level:DEBUG ~timestamp:1234.0 "debug msg";
;;

(******************************************************************************)
